/*----------------------------------------------------------------------------
 *  COMPONENT_NAME: xx
 *
 *  ORIGINS: 27
 *
 *  THIS PRODUCT CONTAINS RESTRICTED MATERIALS OF IBM
 *  5639-D57 (C)  COPYRIGHT International Business Machines Corp. 1997, 1998
 *  All Rights Reserved *  Licensed Materials - Property of IBM
 *  The source code for this program is not published or otherwise divested
 *  of its trade secrets, irrespective of what has been deposited with the
 *  U.S. Copyright Office.
 *
 *# @(#) 1.13 ob/system/helpstub.c, ivb.ob.system, ob.r20.ui 3/31/98 08:25:55 [9/21/98 06:16:04]
 *
 *  DESCRIPTION:
 *  blah blah blah blah blah blah blah blah blah blah blah blah blah
 *
 *  Change History:
 *
 *  Date      Programmer Defect    Description
 *  --------  ---------  ------    -------------------------------------------
 *  01/12/98  yenlu      103732    Workaround for Netscape WWW_GetWindowInfo bug.
 *  01/08/98  yenlu      103382    Improved browser management and native cleanup.
 *  01/05/98  yenlu      103281    Prevent Netscape from trapping.
 *  01/05/98  yenlu      103229    Object Builder managed help windows support.
 *  12/21/97  yenlu      102926    Launching help may trap JRE.
 *  12/18/97  yenlu      102908    IE detection improperly overriden.
 *  12/08/97  yenlu      102004    Support for detecting default browser on NT.
 *  12/03/97  yenlu      101610    Initial creation.
 *----------------------------------------------------------------------------
 */

/* Copyright constant */
static const char copyright[] = "5639-D57 (C)  COPYRIGHT International Business Machines Corp. 1997, 1998";

#include <windows.h>
#include <ddeml.h>
#include <string.h>
#include "clonehelpstub.h"

/* Define macro constants used to handle the user's browser preferences */
#define IESERV                      "iexplore"
#define NNSERV                      "netscape"
#define OPENNEWWINDOW               0
#define REUSEWINDOW                 1
#define INITIAL_BROWSERWINDOW_ID    0xFFFFFFFF
#define LAST_ACTIVEBROWSERWINDOW_ID 0xFFFFFFFF

/* Define private macros for the browser shutting down test */
#define ALIVE     0
#define HCONV_BAD 1
#define HDATA_BAD 2

/* Define global variables:
 * 1) BOOL for whether or not a new browser window was launched.
 * 2) The DDE id for Object Builder's reserved help window.
 * 3) The saved browser window information string.
 */
static BOOL  wasLaunched        = FALSE;
static DWORD dwBrowserWindowId  = INITIAL_BROWSERWINDOW_ID;
static char* savedWindowInfo    = NULL;

/* Define internal function prototypes */

/* The main stub function */
int mainStub(const char*,const char*,const char*,int);

/* Browser initialization functions */
int  initBrowser(DWORD,const char*,const char*,const char*,int);
BOOL shutDownTest(DWORD,const char*);
BOOL launchBrowser(const char*,const char*,HANDLE*);
BOOL setBrowserWindowId(DWORD,const char*,const char*,int);

/* DDE wrappers */
BOOL     ddeInit(DWORD*);
HCONV    ddeConn(DWORD,const char*,const char*);
HDDEDATA ddeClientTransReq(DWORD,HCONV,const char*);
DWORD    ddeActivateBrowser(DWORD,DWORD,const char*);
void     ddeOpenURL(DWORD,DWORD,const char*,const char*,int);
HDDEDATA ddeListWindows(DWORD,const char*);
DWORD    ddeTopFrameAncestor(DWORD,DWORD,const char*);
char*    ddeGetWindowInfo(DWORD,DWORD,const char*);
HDDEDATA CALLBACK DdeCallback(UINT,UINT,HCONV,HSZ,HSZ,HDDEDATA,DWORD,DWORD);

/* Registry accessor */
char* getDefaultValue(HKEY,const char*);

/* URL operators and convenience functions */
char*   URLProtocol(char*);
char*   windowInfoURL(char*);
char*   compactURL(char*,int);
BOOL    compareURLs(const char*,const char*);
char*   obtainString(JNIEnv* jniEnv,jbyteArray jbArray,jint jSize);


/* The native stub for nativeCleanup(). This should clean up any resources
 * held by the native module. This function should only be called when Java
 * VM is about to exit.
 */
JNIEXPORT void JNICALL Java_com_ibm_ivb_cloning_system_HTMLHelpHandler_nativeCleanup
(JNIEnv* jniEnv,jobject jobj)
{
  if (savedWindowInfo != NULL)
  {
    free(savedWindowInfo);
    savedWindowInfo = NULL;
  }
}

/* A function which takes a Java byte array and returns the native char*
 * representation. This helper is needed due to corruption which apparently
 * occurs when the byte array is passed from a Java thread. It is the caller's
 * responsibility to manage the returned buffer's memory space.
 */
char* obtainString(JNIEnv* jniEnv,jbyteArray jbArray,jint jSize)
{
  int i, nativeElemsSize;
  char* nativeElems;
  char* processedString;

  nativeElems = (char*)(*jniEnv)->GetByteArrayElements(jniEnv,jbArray,NULL);
  processedString = (char*)malloc((jSize+1)*sizeof(char));
  memset(processedString,0,jSize+1);
  strncpy(processedString,nativeElems,jSize);
  (*jniEnv)->ReleaseByteArrayElements(jniEnv,jbArray,(jbyte*)nativeElems,0);
  return processedString;
}

/* The native stub for showURLNative(). Java expects a return code of 1 to
 * indicate failure and 0 to indicate success.
 */
JNIEXPORT jint JNICALL Java_com_ibm_ivb_cloning_system_HTMLHelpHandler_showURLNative
(JNIEnv* jniEnv,
 jobject jobj,
 jbyteArray jBrowserPath,
 jint jBrowserPathLength,
 jbyteArray jService,
 jint jServiceLength,
 jbyteArray jUrlName,
 jint jUrlNameLength,
 jint isNewBrowserWindowFlag)
{
  char* browserPath;
  char* service;
  char* urlName;
  int rc;

  browserPath = obtainString(jniEnv,jBrowserPath,jBrowserPathLength);
  service = obtainString(jniEnv,jService,jServiceLength);
  urlName = obtainString(jniEnv,jUrlName,jUrlNameLength);

  /* Call the main method in this stub. */
  rc = mainStub(browserPath,service,urlName,isNewBrowserWindowFlag);

  free(urlName);
  free(service);
  free(browserPath);

  return rc;
}


/* Attempt to launch the browser and grab the DDE browser window id for
 * Object Builder usage.
 */
int initBrowser
(DWORD idInst,
 const char* browserPath,
 const char* service,
 const char* urlName,
 int isNewBrowserWindowFlag)
{
  int rc;
  HANDLE hProcess;

  rc = ((launchBrowser(browserPath,urlName,&hProcess)==TRUE) ? 0 : 1);

  if (rc == 0)
  {
    dwBrowserWindowId = INITIAL_BROWSERWINDOW_ID;
    wasLaunched = TRUE;

    /* Wait for the browser to fully launch. The shutdown test can be
     * used for this.
     *
     * NOTE: IE 4 will exit immediately if urlName does not exist. As a
     *       result, we must also check if the launched browser process
     *       is still alive or not in addition to the shutdown test.
     */
    while (shutDownTest(idInst,service)!=ALIVE)
    {
      DWORD exitCode;
      GetExitCodeProcess(hProcess,&exitCode);
      if (exitCode != STILL_ACTIVE)
        break;
    }

    /* Save the window identification information. */
    setBrowserWindowId(idInst,service,urlName,isNewBrowserWindowFlag);
  }

  return rc;
}


/* A DDE browser service shutdown tester. If a browser is truly shutdown,
 * the conversation handle should become invalid. There is a period of
 * transition during shutdown where the conversation handle will not be
 * invalid although DDE transaction data will be invalid. If this is not
 * checked, browser services could crash. The function returns the following:
 *
 * HDATA_BAD - DDE transaction data is bad.
 * HCONV_BAD - DDE conversation handle is invalid.
 * ALIVE     - the service is alive and well.
 */
int shutDownTest(DWORD idInst,const char* service)
{
  HCONV hConv;
  HDDEDATA hData;
  BOOL bRc;

  hConv = ddeConn(idInst,service,"WWW_ListWindows");

  if (hConv != 0L)
  {
    hData = ddeClientTransReq(idInst,hConv,service);
    if (hData == 0)
      return HDATA_BAD;
    else
      return ALIVE;
  }
  else
    return HCONV_BAD;
}


/* The main stub for this library. It performs the following sequence of tasks:
 * 1) attempt to register this module with the DDE Management Library (DDEML).
 * 2) attempt to open a conversation with a running browser to activate it.
 * 3) attempt to activate the browser.
 * 4) attempt to open a conversation with a running browser to change its URL.
 * 5) attempt to open the browser at a new URL.
 * 6) disconnect the DDE conversation.
 * 7) Unitialize the DDEML.
 */
int mainStub
(const char* browserPath,
 const char* service,
 const char* urlName,
 int isNewBrowserWindowFlag)
{
  DWORD idInst = 0;
  int rc;
  BOOL bRc;
  HCONV hConv = 0L;

  /* Attempt to register this module with the DDE Management Library (DDEML) */
  if (!ddeInit(&idInst))
    return 1;

  /* Check for the an executing instance of the user specified or default
   * running browser. If no instance is detected then continue calling
   * shutDownTest() until the conversation handle becomes invalid (i.e. HCONV_BAD)
   * . This signals complete browser shutdown. We can then safely launch a new
   * browser.
   */
  if (shutDownTest(idInst,service) != ALIVE)
  {
    while (shutDownTest(idInst,service) == HDATA_BAD);
    rc = initBrowser(idInst,browserPath,service,urlName,isNewBrowserWindowFlag);
    DdeUninitialize(idInst);
    return rc;
  }

  /* A running browser has responded. The Object Builder help's DDE browser window
   * id should be updated (in case it was closed previously).
   */
  setBrowserWindowId(idInst,service,urlName,isNewBrowserWindowFlag);

  /* Open the URL in the browser. */
  ddeOpenURL(dwBrowserWindowId,idInst,service,urlName,isNewBrowserWindowFlag);

  /* Activate the browser window so that it gets the focus and moves to the front
   * of the window Z-order if possible.
   */
  if (dwBrowserWindowId != 0)
    ddeActivateBrowser(dwBrowserWindowId,idInst,service);

  DdeUninitialize(idInst);

  rc = 0;

  return rc;
}


/* The native stub for showURLNativeDefault. This function attempts to launch
 * the system's default browser based on registry settings and is typically
 * called by HTMLHelpHandler if the user-specified browser and path settings fail.
 */
JNIEXPORT jint JNICALL Java_com_ibm_ivb_cloning_system_HTMLHelpHandler_showURLNativeDefault
(JNIEnv* jniEnv,
 jobject jobj,
 jbyteArray jUrlName,
 jint jUrlNameLength,
 jint isNewBrowserWindowFlag)
{
  static char baseKey[] = "SOFTWARE\\Classes\\";
  static char baseCommandKey[] = "\\shell\\open\\command";
  static char baseDdeKey[] = "\\shell\\open\\ddeexec\\Application";
  char* defBrsAppKey;
  char* queryKey;
  char* brsExec;
  char* ddeService;
  char* urlName;
  int queryKeySize;
  int rc;

  /* Find the default browser's application key <defBrsAppKey> */

  /* Form queryKey = SOFTWARE\Classes\.html */
  queryKeySize = strlen(baseKey) +
                 strlen(".html") +
                 1;
  queryKey = (char*)malloc(queryKeySize*sizeof(char));
  strcpy(queryKey,baseKey);
  strcat(queryKey,".html");

  defBrsAppKey = getDefaultValue(HKEY_LOCAL_MACHINE,queryKey);

  free(queryKey);

  if (defBrsAppKey == NULL)
    return 1;

  /* Find the default browser's fully-qualified executable <brsExec> */

  /* Form queryKey = SOFTWARE\Classes\<defBrsAppKey>\shell\open\command */
  queryKeySize = strlen(baseKey) +
                 strlen(defBrsAppKey) +
                 strlen(baseCommandKey) +
                 1;
  queryKey = (char*)malloc(queryKeySize*sizeof(char));
  strcpy(queryKey,baseKey);
  strcat(queryKey,defBrsAppKey);
  strcat(queryKey,baseCommandKey);

  brsExec = getDefaultValue(HKEY_LOCAL_MACHINE,queryKey);

  free(queryKey);

  if (brsExec == NULL)
  {
    free(defBrsAppKey);
    return 1;
  }

  /* Find the default browser's DDE service name

  /* Form queryKey = SOFTWARE\Classes\<defBrsAppKey>\shell\open\ddeexec\Application */
  queryKeySize = strlen(baseKey) +
                 strlen(defBrsAppKey) +
                 strlen(baseDdeKey) +
                 1;
  queryKey = (char*)malloc(queryKeySize*sizeof(char));
  strcpy(queryKey,baseKey);
  strcat(queryKey,defBrsAppKey);
  strcat(queryKey,baseDdeKey);

  ddeService = getDefaultValue(HKEY_LOCAL_MACHINE,queryKey);

  free(queryKey);

  if (ddeService == NULL)
  {
    free(brsExec);
    free(defBrsAppKey);
    return 1;
  }

  else if (stricmp(ddeService,IESERV))
  {
    /* For Netscape, the DDE service name registered with the registry is
     * "NSShell". "netscape" should be used explicitly in place of the
     * registered name.
     */
    free(ddeService);
    ddeService = (char*)malloc((strlen(NNSERV)+1)*sizeof(char));
    strcpy(ddeService,NNSERV);
  }

  /* Attempt to show help with the information we have */
  urlName = obtainString(jniEnv,jUrlName,jUrlNameLength);
  rc = mainStub(brsExec,ddeService,urlName,isNewBrowserWindowFlag);
  free(urlName);

  free(ddeService);
  free(brsExec);
  free(defBrsAppKey);

  return rc;
}


/* Get the default data value for a given registry entry. Management of the
 * returned data is then the user's responsibility.
 */
char* getDefaultValue(HKEY hKeyRoot,const char* subKey)
{
  HKEY hKey;
  LONG rc;
  DWORD bufferSize = 10;
  char* result;

  /* Open the Registry */
  rc = RegOpenKeyEx(hKeyRoot,subKey,0,KEY_READ,&hKey);

  if (rc != ERROR_SUCCESS)
    return NULL;

  /* Query the opened key's default value */
  result = (char*)malloc(bufferSize*sizeof(char));
  rc = RegQueryValueEx(hKey,"",NULL,NULL,result,&bufferSize);

  if (rc == ERROR_MORE_DATA)
  {
    /* The buffer wasn't large enough. Re-query with a larger buffer. */
    result = (char*)realloc((void*)result,bufferSize);
    rc = RegQueryValueEx(hKey,"",NULL,NULL,result,&bufferSize);
  }

  if (rc != ERROR_SUCCESS)
  {
    RegCloseKey(hKey);
    free(result);
    return NULL;
  }

  RegCloseKey(hKey);

  return result;
}


/* The DDE callback function */
HDDEDATA CALLBACK DdeCallback(UINT iType,
                              UINT iFmt,
                              HCONV hConv,
                              HSZ hsz1,
                              HSZ hsz2,
                              HDDEDATA hData,
                              DWORD dwData1,
                              DWORD dwData2)
{
  return NULL;
}


/* Initialize this client module with the DDE Management Library */
BOOL ddeInit(DWORD* pidInst)
{
  UINT rc;

  rc = DdeInitialize(pidInst,
                     (PFNCALLBACK)&DdeCallback,
                     APPCLASS_STANDARD|APPCMD_CLIENTONLY,
                     0L);

  if (rc != DMLERR_NO_ERROR)
    return FALSE;

  return TRUE;
}


/* Connect to the DDE Server (a browser in this case) */
HCONV ddeConn(DWORD idInst,const char* szService,const char* szTopic)
{
  HSZ hszService,hszTopic;
  HCONV hConv;

  hszService = DdeCreateStringHandle(idInst,szService,CP_WINANSI);
  hszTopic   = DdeCreateStringHandle(idInst,szTopic,CP_WINANSI);

  hConv = DdeConnect(idInst,hszService,hszTopic,NULL);

  DdeFreeStringHandle(idInst,hszService);
  DdeFreeStringHandle(idInst,hszTopic);

  return hConv;
}


/* Make a DDE Request Transaction to the server. Any transaction data will be
 * returned via the HDDEDATA handle. The caller is then responsible for managing
 * the handle.
 */
HDDEDATA ddeClientTransReq(DWORD idInst,HCONV hConv,const char* szItem)
{
  HSZ hszItem;
  HDDEDATA hData;

  hszItem = DdeCreateStringHandle(idInst,szItem,CP_WINANSI);
  hData = DdeClientTransaction(NULL,0,hConv,hszItem,CF_TEXT,XTYP_REQUEST,3000,NULL);
  DdeFreeStringHandle(idInst,hszItem);

  return hData;
}


/* Launch the browser on some webpage */
BOOL launchBrowser(const char* browserPath,const char* urlName,HANDLE* pHprocess)
{
  char* command = NULL;
  char* args[]={(char*)urlName};
  STARTUPINFO startInfo;
  PROCESS_INFORMATION procInfo;
  BOOL rc;
  int cSize;

  startInfo.cb = sizeof(STARTUPINFO);
  startInfo.lpReserved = NULL;
  startInfo.lpDesktop = NULL;
  startInfo.lpTitle = NULL;
  startInfo.dwFlags = STARTF_USESHOWWINDOW;
  startInfo.wShowWindow = SW_SHOWNORMAL;
  startInfo.cbReserved2 = 0;
  startInfo.lpReserved2 = NULL;

  /* This function also needs to support the possibility that browserPath may
   * contain some "%#" argument parameters. If Netscape's open command is
   * retried via the registry, we will encounter such a case. In this case,
   * the argument can be filled in with FormatMessage.
   */

  cSize = FormatMessage(FORMAT_MESSAGE_ARGUMENT_ARRAY |
                        FORMAT_MESSAGE_ALLOCATE_BUFFER |
                        FORMAT_MESSAGE_FROM_STRING |
                        FORMAT_MESSAGE_MAX_WIDTH_MASK,
                        browserPath,
                        0,
                        0,
                        (LPTSTR)&command,
                        0,
                        args);

  if (cSize == strlen(browserPath))
  {
    /* The browserPath variable contained no format parameters. Consequently,
     * the URL argument must be appended to the browserPath before launching.
     * NOTE: The buffer allocated by FormatMessage() may not be LocalReAlloc-
     * able. Consequently, it should be freed and then rebuilt with the correct
     * size and contents.
     */
    LocalFree(command);
    cSize = strlen(browserPath) + strlen(" ") + strlen(urlName) + 1;
    command = (char*)LocalAlloc(LPTR,cSize);

    strcpy(command,browserPath);
    strcat(command," ");
    strcat(command,urlName);
  }

  rc = CreateProcess(NULL,
                     command,
                     NULL,
                     NULL,
                     FALSE,
                     NORMAL_PRIORITY_CLASS | CREATE_NEW_PROCESS_GROUP,
                     NULL,
                     ".",
                     &startInfo,
                     &procInfo);

  LocalFree(command);

  *pHprocess = procInfo.hProcess;
  WaitForInputIdle(procInfo.hProcess,INFINITE);

  return rc;
}


/* Activate the browser with DDE id dwWindowId via DDE transaction. The return
 * value is the window id returned by the server (0 to indicate failure).
 */
DWORD ddeActivateBrowser(DWORD dwWindowId,DWORD idInst,const char* service)
{
  static char szItemBase[]=",0x0";
  char szWindowId[33];
  int szItemSize;
  HDDEDATA hData;
  char* szItem;
  DWORD dwSize;
  DWORD dwActivatedWindowId = 0x0;
  HCONV hConv;

  /* Open a DDE conversation for topic: WWW_Activate */
  hConv = ddeConn(idInst,service,"WWW_Activate");

  if (hConv == 0L)
    return dwActivatedWindowId;

  /* Form the DDE Topic Item for WWW_Activate. This will be of the form:
   * "dwWindowId,0x0".
   */
  itoa(dwWindowId,szWindowId,10);
  szItemSize = strlen(szWindowId) + strlen(szItemBase) + 1;
  szItem = (char*)malloc(szItemSize*sizeof(char));
  strcpy(szItem,szWindowId);
  strcat(szItem,szItemBase);

  /* Make the transaction with the server. The browser with window id windowId
   * should grab the focus.
   */
  hData = ddeClientTransReq(idInst,hConv,szItem);

  /* Close the DDE conversation now that the transaction has been performed. */
  DdeDisconnect(hConv);

  if (hData != 0)
  {
    dwActivatedWindowId = (DWORD)(*(DdeAccessData(hData,&dwSize)));
    DdeUnaccessData(hData);
    DdeFreeDataHandle(hData);
  }

  free(szItem);

  return dwActivatedWindowId;
}


/* Perform a WWW_OpenURL transaction with the DDE server */
void ddeOpenURL(DWORD dwWindowId,
                DWORD idInst,
                const char* service,
                const char* urlName,
                int isNewBrowserWindowFlag)
{
  static char szItemBaseFront[] = ",,";
  static char szItemBaseBack[] = ",0x03,,,";
  char szWindowId[33];
  HDDEDATA hData;
  int szItemSize;
  char* szItem;
  HCONV hConv;
  int count = 0;

  /* Open a DDE conversation for topic: WWW_OpenURL. */
  hConv = ddeConn(idInst,service,"WWW_OpenURL");

  /* Form the DDE Topic Item for WWW_OpenURL. This will be of the form:
   * "urlName,,dwWindowId,0x03,,,".
   */
  itoa(dwWindowId,szWindowId,10);

  szItemSize = strlen(urlName) +
               strlen(szItemBaseFront) +
               strlen(szWindowId) +
               strlen(szItemBaseBack) +
               1;
  szItem = (char*)malloc(szItemSize*sizeof(char));
  strcpy(szItem,urlName);
  strcat(szItem,szItemBaseFront);
  strcat(szItem,szWindowId);
  strcat(szItem,szItemBaseBack);

  /* Make the transaction with the server. The topic should cause the
   * running browser to open the URL given by urlName.
   */

  hData = ddeClientTransReq(idInst,hConv,szItem);

  /* Close the DDE conversation now that the transaction has been performed. */
  DdeDisconnect(hConv);

  /* Get the window identification data (window id and window info) which now
   * displays the URL. This is only necessary for the "open a separate browser
   * window" case.
   */
  if (hData != 0 && isNewBrowserWindowFlag == OPENNEWWINDOW)
  {
    if (savedWindowInfo != NULL)
    {
      free(savedWindowInfo);
      savedWindowInfo = NULL;
    }

    if (dwBrowserWindowId == 0x0)
    {
      /* Get the newest window. It should be the one with the largest DDE
       * window id.
       */
      DWORD* windowIdList;
      DWORD maxWindowId;
      DWORD dwSize;
      int count;

      hData = ddeListWindows(idInst,service);

      if (hData != 0)
      {
        windowIdList = (DWORD*)DdeAccessData(hData,&dwSize);

        maxWindowId = windowIdList[0];
        count = 1;

        while (windowIdList[count] != 0x0 && windowIdList[count] != LAST_ACTIVEBROWSERWINDOW_ID)
        {
          if (windowIdList[count] > maxWindowId)
            maxWindowId = windowIdList[count];
          count++;
        };

        dwBrowserWindowId = maxWindowId;

        DdeUnaccessData(hData);
        DdeFreeDataHandle(hData);
      }
    }

    /* To get valid window info, the browser must first finish loading the URL.
     * This is an asynchronous operation so we can continue polling the window
     * information until it contains the desired URL. Try 100 times before
     * giving up.
     */
    do
    {
      char* loadedURL;
      savedWindowInfo = ddeGetWindowInfo(dwBrowserWindowId,idInst,service);

      if (savedWindowInfo == NULL)
        break;

      loadedURL = windowInfoURL(savedWindowInfo);

      if (compareURLs(loadedURL,urlName))
      {
        free(loadedURL);
        break;
      }

      free(loadedURL);
      free(savedWindowInfo);
      savedWindowInfo = NULL;
      count++;
    } while (count < 100);

    if (count == 100 || savedWindowInfo == NULL)
      dwBrowserWindowId = 0x0;

    wasLaunched = FALSE;

  }

  free(szItem);
}


/* Given a window or frame id, find the top-most parent frame. This allows
 * Object Builder help to overwrite the browser's main window rather than
 * just a frame. This is implemented via a recursive procedure which makes
 * a DDE conversation for WWW_GetFrameParent until the result is NULL. When
 * the top ancestor id is explained, the client transaction with the id
 * should return NULL.
 */
DWORD ddeTopFrameAncestor(DWORD dwFrameId,DWORD idInst,const char* service)
{
  HCONV hConv;
  HDDEDATA hData;
  char szFrameId[33];
  DWORD dwSize;

  itoa(dwFrameId,szFrameId,10);

  /* Open a conversation handle for topic WWW_GetFrameParent */
  hConv = ddeConn(idInst,service,"WWW_GetFrameParent");

  /* Make the transaction with the server. The server should return the id of
   * the parent of the current frame. If the return value is NULL, then the
   * current frame is the parent.
   */
  hData = ddeClientTransReq(idInst,hConv,szFrameId);

  DdeDisconnect(hConv);

  if (hData != NULL)
  {
    /* Continue the DDE call on the parent frame to find its parent */
    DWORD parentFrameId = (DWORD)(*(DdeAccessData(hData,&dwSize)));
    DdeUnaccessData(hData);
    DdeFreeDataHandle(hData);
    return ddeTopFrameAncestor(parentFrameId,idInst,service);
  }

  return dwFrameId;
}


/* Obtain the window information for the last active browser window. This is
 * accomplished via* a DDE conversation with the WWW_GetWindowInfo topic. A
 * pointer to the window information will be returned and then it is the
 * responsibility of the caller to manage this pointer. If the DDE calls fail
 * (typically if all browser windows are shutdown before the transaction),
 * NULL is returned.
 *
 * SPECIAL NOTE: There appears to be a bug with WWW_GetWindowInfo on Netscape.
 *               When this DDE topic is used with an invalid window id or a 0x0
 *               window id, Netscape will crash. As a result, we should only be
 *               using LAST_ACTIVEBROWSERWINDOW_ID for this.
 *
 * To accomodate the netscape bug, the following initialization steps will be
 * taken:
 *
 * 1) find the top frame ancestor for the dwWindowId parameter.
 * 2) activate the top frame ancestor from 1 so that it becomes the last active
 *    frame.
 * 3) the window information is then queried from the last active frame.
 */
char* ddeGetWindowInfo(DWORD dwWindowId,DWORD idInst,const char* service)
{
  HCONV hConv;
  HDDEDATA hData;
  DWORD dwParentFrameId;
  char szWindowId[33];
  DWORD dwSize;
  char* windowInfo;

  /* Activate the top frame ancestor for dwWindowId so that it becomes the
   * last active frame.
   */
  dwParentFrameId = ddeTopFrameAncestor(dwWindowId,idInst,service);
  ddeActivateBrowser(dwParentFrameId,idInst,service);

  /* Even though the correct browser window id should have been activated,
   * Internet Explorer V4.0 may sometimes give the information for the
   * wrong window if we use windowid = LAST_ACTIVEBROWSERWINDOW_ID. The
   * fix is to always explicitly use the correct window id for IE 4.0.
   */
  if (stricmp(service,IESERV)==0)
    itoa(dwParentFrameId,szWindowId,10);
  else
    itoa(LAST_ACTIVEBROWSERWINDOW_ID,szWindowId,10);

  /* Open a conversation handle for topic WWW_GetWindowInfo */
  hConv = ddeConn(idInst,service,"WWW_GetWindowInfo");

  if (hConv == 0L)
    return NULL;

  /* Make the transaction with the server. The server should return the window
   * info for the DDE window id dwWindowId. A NULL return value implies that the
   * transaction has failed.
   */
  hData = ddeClientTransReq(idInst,hConv,szWindowId);

  DdeDisconnect(hConv);

  if (hData != NULL)
  {
    /* Get the required size for the windowInfo */
    char* s = DdeAccessData(hData,&dwSize);

    /* Allocate the windowInfo buffer to receive the windowinfo. */
    windowInfo = (char*)malloc(dwSize*sizeof(char));
    strcpy(windowInfo,s);
    DdeUnaccessData(hData);
    DdeFreeDataHandle(hData);
  }
  else
    windowInfo = NULL;

  return windowInfo;
}


/* Get a list of available browser windows via DDE topic WWW_ListWindows.
 *
 * SPECIAL NOTE: To properly detect when a browser and all its windows have
 *               shut down, we must check for 0 data returned as well as an
 *               invalid conversation handle.
 */
HDDEDATA ddeListWindows(DWORD idInst,const char* service)
{
  HCONV hConv;
  HDDEDATA hData;
  static char szItem[] = "WWW_ListWindows";

  hConv = ddeConn(idInst,service,szItem);

  if (hConv == 0L)
    return 0;

  hData = ddeClientTransReq(idInst,hConv,szItem);

  DdeDisconnect(hConv);

  return hData;
}


/* Refresh the help window's DDE id from the list of available browser windows.
 * The function returns FALSE if it cannot get the Object Builder Help Window Id
 * via DDE calls.
 */
BOOL setBrowserWindowId
(DWORD idInst,
 const char* service,
 const char* urlName,
 int isNewBrowserWindowFlag)
{
  HDDEDATA hData;
  DWORD dwSize,dwCandidateId,dwActivatedId;
  char* szCandidateInfo;
  DWORD* windowIdList;
  BOOL found = FALSE;
  int count = 0;

  if (isNewBrowserWindowFlag == REUSEWINDOW)
  {
    /* The user has opted to not open a new browser window for Object Builder
     * help. In this case, the parent window of the last active frame is used
     * to display help. The procedure for determining this is as follows:
     *
     * 1) Get the DDE window id of the last active browser window/frame.
     * 2) Get the topmost frame parent of 1 and set this to be the window id.
     */
    dwActivatedId = ddeActivateBrowser(LAST_ACTIVEBROWSERWINDOW_ID,idInst,service);
    dwBrowserWindowId = ddeTopFrameAncestor(dwActivatedId,idInst,service);
    if (wasLaunched)
      wasLaunched = FALSE;
    return TRUE;
  }

  /* The user has opted to assign a browser window specifically for Object
   * Builder. If this window is already available and unmodified amongst
   * the list of currently available browser windows, do nothing. Otherwise,
   * set the browser window id such that a new window will be created. 0x0
   * serves this purpose.
   */
  hData = ddeListWindows(idInst,service);

  if (hData == 0)
  {
    dwBrowserWindowId = INITIAL_BROWSERWINDOW_ID;
    return FALSE;
  }

  windowIdList = (DWORD*)DdeAccessData(hData,&dwSize);

  /* If the browser was just launched, check if this is the first ever launch
   * and then assign the window id.
   */
  if (wasLaunched)
  {
    if (savedWindowInfo != NULL)
    {
      free(savedWindowInfo);
      savedWindowInfo = NULL;
    }

    dwBrowserWindowId = ddeTopFrameAncestor(windowIdList[0],idInst,service);

    count = 0;

    do
    {
      /* Attempt to get the window information for the browser window that was
       * just launched. If the user has just quickly shut down the browser, no
       * window information will be saved. When valid window information is
       * obtained, it must be checked that it does indeed correspond to the
       * URL we want loaded. Do this for a maximum of 100 times to prevent an
       * invalid URLs from hanging the system.
       */
      char* infoURL;

      savedWindowInfo = ddeGetWindowInfo(dwBrowserWindowId,idInst,service);

      if (savedWindowInfo == NULL)
        break;

      infoURL = windowInfoURL(savedWindowInfo);

      if (compareURLs(infoURL,urlName))
      {
        free(infoURL);
        break;
      }

      free(infoURL);
      free(savedWindowInfo);

      count++;

    } while (count < 100);

    wasLaunched = FALSE;
    DdeUnaccessData(hData);
    DdeFreeDataHandle(hData);

    return TRUE;
  }

  /* The browser was not launched but there is no savedWindowInfo (likely the
   * very first time help is invoked and browser windows are up. The Help
   * Manager must assume that none of the windows belong to it and, as a result,
   * should launch a new window.
   */
  if (savedWindowInfo == NULL)
  {
    dwBrowserWindowId = 0x0;
    DdeUnaccessData(hData);
    DdeFreeDataHandle(hData);
    return TRUE;
  }

  /* Attempt to find the Object Builder help window amongst the list of windows.
   * The Object Builder help window should be the first window in the list whose
   * DDE windowInfo matches the saved windowInfo. If this can't be found, the
   * window id should be set to 0x0 to open a new browser window.
   *
   * Since savedWindowInfo is not NULL, we should have a valid dwBrowserWindowId.
   */

  dwCandidateId = ddeTopFrameAncestor(dwBrowserWindowId,idInst,service);
  szCandidateInfo = ddeGetWindowInfo(dwCandidateId,idInst,service);

  if (szCandidateInfo != NULL)
  {
    if (strcmp(szCandidateInfo,savedWindowInfo)==0)
    {
      dwBrowserWindowId = dwCandidateId;
      free(szCandidateInfo);
      found = TRUE;
    }
  }

  if (!found)
    dwBrowserWindowId = 0x0;

  DdeUnaccessData(hData);
  DdeFreeDataHandle(hData);

  return TRUE;

}


/* Given the results from ddeGetWindowInfo, this function extracts the URL.
 * For both Netscape and Internet Explorer, the URL information is the first
 * string enclosed in quotes in the window info. The user is responsible for
 * managing the returned data.
 */
char* windowInfoURL(char* windowInfo)
{
  char* url;
  char* temp;
  int size;
  int count = 0;

  if (windowInfo == NULL)
    return NULL;

  temp = &windowInfo[1];
  while (temp[count] != '\"' && count < strlen(temp)) count++;

  size = count+1;
  url = (char*)malloc(size*sizeof(char));
  memset(url,0,size);

  strncpy(url,temp,size-1);
  url[size-1]='\0';

  return url;
}


/* Given a URL, this function extracts the protocol along with the colon
 * separator. The user is responsible for managing the returned data.
 */
char* URLProtocol(char* url)
{
  char* protocol;
  int size;
  int count = 0;

  while (url[count] != ':' && count < strlen(url)) count++;

  size = count + 2;
  protocol = (char*)malloc(size*sizeof(char));
  memset(protocol,0,size);
  strncpy(protocol,url,size-1);
  protocol[size-1]='\0';
  return protocol;
}


/* Compact a given URL with a protocol of protocolLength. This function
 * will generate the section of the URL which does not include the protocol,
 * colon and any leading '/'.
 *
 * SPECIAL NOTE:
 * DO NOT DELETE the return value. It is not newly created memory!
 */
char* compactURL(char* url,int protocolLength)
{
  char* compactedURL;
  int count;

  /* Skip the protocol and colon */
  compactedURL = &url[protocolLength];

  /* Skip leading '/' */
  while (*compactedURL == '/')
    compactedURL = &compactedURL[1];

  count = 0;

  /* Consistency conversions:
   * 1) Convert "\" to "/"
   * 2) Convert "|" to ":"
   */
  do
  {
    if (compactedURL[count] == '\\')
      compactedURL[count] = '/';
    else if (compactedURL[count] == '|')
      compactedURL[count] = ':';
    count++;
  } while (count < strlen(compactedURL));

  return compactedURL;
}


/* Compare two URLs. The protocols are extracted and compared first and then
 * the URL is compared after compaction.
 */
BOOL compareURLs(const char* url1,const char* url2)
{
  /* 1) extract the protocol.
   * 2) remove any leading '/' characters following the protocol.
   * 3) extract the rest of the string and replace '\' with '/'.
   * 4) case-insensitive compare results for 2 and 4.
   */
  BOOL rc = FALSE;

  char* protocol1;
  char* protocol2;

  char* urlCopy1;
  char* urlCopy2;

  urlCopy1 = (char*)malloc((strlen(url1)+1)*sizeof(char));
  urlCopy2 = (char*)malloc((strlen(url2)+1)*sizeof(char));

  strcpy(urlCopy1,url1);
  strcpy(urlCopy2,url2);

  protocol1 = URLProtocol(urlCopy1);
  protocol2 = URLProtocol(urlCopy2);

  if (stricmp(protocol1,protocol2) != 0)
    rc = FALSE;
  else
  {
    char* urlCopyCompacted1 = compactURL(urlCopy1,strlen(protocol1));
    char* urlCopyCompacted2 = compactURL(urlCopy2,strlen(protocol2));
    if (stricmp(urlCopyCompacted1,urlCopyCompacted2) != 0)
      rc = FALSE;
    else
      rc = TRUE;
  }

  free(protocol2);
  free(protocol1);
  free(urlCopy2);
  free(urlCopy1);

  return rc;
}
